import logging
import sys

from celery import shared_task
from celery.exceptions import Ignore, Reject
from dateutil import parser
from django.core.cache import cache
from fitbit.exceptions import HTTPTooManyRequests

from . import utils
from .models import UserFitbit, TimeSeriesData, TimeSeriesDataType


logger = logging.getLogger(__name__)
LOCK_EXPIRE = 60 * 5 # Lock expires in 5 minutes


@shared_task
def subscribe(fitbit_user, subscriber_id):
    """ Subscribe to the user's fitbit data """

    fbusers = UserFitbit.objects.filter(fitbit_user=fitbit_user)
    for fbuser in fbusers:
        fb = utils.create_fitbit(**fbuser.get_user_data())
        try:
            fb.subscription(fbuser.user.id, subscriber_id)
        except:
            exc = sys.exc_info()[1]
            logger.exception("Error subscribing user: %s" % exc)
            raise Reject(exc, requeue=False)


@shared_task
def unsubscribe(*args, **kwargs):
    """ Unsubscribe from a user's fitbit data """

    fb = utils.create_fitbit(**kwargs)
    try:
        for sub in fb.list_subscriptions()['apiSubscriptions']:
            if sub['ownerId'] == kwargs['user_id']:
                fb.subscription(sub['subscriptionId'], sub['subscriberId'],
                                method="DELETE")
    except:
        exc = sys.exc_info()[1]
        logger.exception("Error unsubscribing user: %s" % exc)
        raise Reject(exc, requeue=False)



@shared_task
def get_time_series_data(fitbit_user, cat, resource, date=None):
    """ Get the user's time series data """

    try:
        _type = TimeSeriesDataType.objects.get(category=cat, resource=resource)
    except TimeSeriesDataType.DoesNotExist:
        logger.exception("The resource %s in category %s doesn't exist" % (
            resource, cat))
        raise Reject(sys.exc_info()[1], requeue=False)

    # Create a lock so we don't try to run the same task multiple times
    sdat = date.strftime('%Y-%m-%d') if date else 'ALL'
    lock_id = '{0}-lock-{1}-{2}-{3}'.format(__name__, fitbit_user, _type, sdat)
    if not cache.add(lock_id, 'true', LOCK_EXPIRE):
        logger.debug('Already retrieving %s data for date %s, user %s' % (
            _type, fitbit_user, sdat))
        raise Ignore()

    fbusers = UserFitbit.objects.filter(fitbit_user=fitbit_user)
    dates = {'base_date': 'today', 'period': 'max'}
    if date:
        dates = {'base_date': date, 'end_date': date}
    try:
        for fbuser in fbusers:
            data = utils.get_fitbit_data(fbuser, _type, **dates)
            for datum in data:
                # Create new record or update existing record
                date = parser.parse(datum['dateTime'])
                tsd, created = TimeSeriesData.objects.get_or_create(
                    user=fbuser.user, resource_type=_type, date=date)
                tsd.value = datum['value']
                tsd.save()
        # Release the lock
        cache.delete(lock_id)
    except HTTPTooManyRequests:
        # We have hit the rate limit for the user, retry when it's reset,
        # according to the reply from the failing API call
        e = sys.exc_info()[1]
        logger.debug('Rate limit reached, will try again in %s seconds' %
                     e.retry_after_secs)
        raise get_time_series_data.retry(e, countdown=e.retry_after_secs)
    except Exception:
        exc = sys.exc_info()[1]
        logger.exception("Exception updating data: %s" % exc)
        raise Reject(exc, requeue=False)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from fitbit import Fitbit

from . import defaults
from .models import UserFitbit


def create_fitbit(consumer_key=None, consumer_secret=None, **kwargs):
    """Shortcut to create a Fitbit instance.

    If consumer_key or consumer_secret are not provided, then the values
    specified in settings are used.
    """
    if consumer_key is None:
        consumer_key = get_setting('FITAPP_CONSUMER_KEY')
    if consumer_secret is None:
        consumer_secret = get_setting('FITAPP_CONSUMER_SECRET')

    if consumer_key is None or consumer_secret is None:
        raise ImproperlyConfigured("Consumer key and consumer secret cannot "
                "be null, and must be explicitly specified or set in your "
                "Django settings")

    return Fitbit(consumer_key, consumer_secret, **kwargs)


def is_integrated(user):
    """Returns ``True`` if we have Oauth info for the user.

    This does not require that the token and secret are valid.

    :param user: A Django User.
    """
    if user.is_authenticated() and user.is_active:
        return UserFitbit.objects.filter(user=user).exists()
    return False


def get_valid_periods():
    """Returns list of periods for which one may request time series data."""
    return ['1d', '7d', '30d', '1w', '1m', '3m', '6m', '1y', 'max']


def get_fitbit_data(fbuser, resource_type, base_date=None, period=None,
        end_date=None):
    """Creates a Fitbit API instance and retrieves step data for the period.

    Several exceptions may be thrown:
        TypeError           - Either end_date or period must be specified, but
                              not both.
        ValueError          - Invalid argument formats.
        HTTPUnauthorized    - 401 - fbuser has bad authentication credentials.
        HTTPForbidden       - 403 - This isn't specified by Fitbit, but does
                                 appear in the Python Fitbit library.
        HTTPNotFound        - 404 - The specific resource doesn't exist.
        HTTPConflict        - 409 - HTTP conflict
        HTTPTooManyRequests - 429 - Hitting the rate limit
        HTTPServerError     - >=500 - Fitbit server error or maintenance.
        HTTPBadRequest      - >=400 - Bad request.
    """
    fb = create_fitbit(**fbuser.get_user_data())
    resource_path = resource_type.path()
    data = fb.time_series(resource_path, user_id=fbuser.fitbit_user,
                          period=period, base_date=base_date,
                          end_date=end_date)
    return data[resource_path.replace('/', '-')]


def get_setting(name, use_defaults=True):
    """Retrieves the specified setting from the settings file.

    If the setting is not found and use_defaults is True, then the default
    value specified in defaults.py is used. Otherwise, we raise an
    ImproperlyConfigured exception for the setting.
    """
    if hasattr(settings, name):
        return getattr(settings, name)
    if use_defaults:
        if hasattr(defaults, name):
            return getattr(defaults, name)
    msg = "{0} must be specified in your settings".format(name)
    raise ImproperlyConfigured(msg)



from django.conf import settings
from django.db import models
from django.utils.encoding import python_2_unicode_compatible


UserModel = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')


@python_2_unicode_compatible
class UserFitbit(models.Model):
    user = models.OneToOneField(UserModel)
    fitbit_user = models.CharField(max_length=32, unique=True)
    auth_token = models.TextField()
    auth_secret = models.TextField()

    def __str__(self):
        return self.user.__str__()

    def get_user_data(self):
        return {
            'resource_owner_key': self.auth_token,
            'resource_owner_secret': self.auth_secret,
            'user_id': self.fitbit_user,
        }


class TimeSeriesDataType(models.Model):
    """
    This model is intended to store information about Fitbit's time series
    resources, which can be found here:
    https://wiki.fitbit.com/display/API/API-Get-Time-Series
    """

    foods = 0
    activities = 1
    sleep = 2
    body = 3
    CATEGORY_CHOICES = (
        (foods, 'foods'),
        (activities, 'activities'),
        (sleep, 'sleep'),
        (body, 'body'),
    )
    category = models.IntegerField(choices=CATEGORY_CHOICES)
    resource = models.CharField(max_length=128)

    def __str__(self):
        return self.path()

    class Meta:
        unique_together = ('category', 'resource',)
        ordering = ['category', 'resource']

    def path(self):
        return '/'.join([self.get_category_display(), self.resource])


class TimeSeriesData(models.Model):
    """
    The purpose of this model is to store Fitbit user data obtained from their
    time series API (https://wiki.fitbit.com/display/API/API-Get-Time-Series).
    """

    user = models.ForeignKey(UserModel)
    resource_type = models.ForeignKey(TimeSeriesDataType)
    date = models.DateField()
    value = models.CharField(null=True, default=None, max_length=32)

    class Meta:
        unique_together = ('user', 'resource_type', 'date')

    def string_date(self):
        return self.date.strftime('%Y-%m-%d')



from __future__ import absolute_import

import celery
import json
import sys

from dateutil import parser
from django.core.cache import cache
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.urlresolvers import reverse
from django.test.utils import override_settings
from freezegun import freeze_time
from mock import MagicMock, patch

from fitbit import exceptions as fitbit_exceptions
from fitbit.api import Fitbit

from fitapp import utils
from fitapp.models import UserFitbit, TimeSeriesData, TimeSeriesDataType
from fitapp.tasks import get_time_series_data

try:
    from io import BytesIO
except ImportError:  # Python 2.x fallback
    from StringIO import StringIO as BytesIO

from .base import FitappTestBase


class TestRetrievalUtility(FitappTestBase):
    """Tests for the get_fitbit_data utility function."""

    def setUp(self):
        super(TestRetrievalUtility, self).setUp()
        self.period = '30d'
        self.base_date = '2012-06-01'
        self.end_date = None

    @patch.object(Fitbit, 'time_series')
    def _mock_time_series(self, time_series=None, error=None, response=None,
                          error_attrs={}):
        if error:
            exc = error(self._error_response())
            for k, v in error_attrs.items():
                setattr(exc, k, v)
            time_series.side_effect = exc
        elif response:
            time_series.return_value = response
        resource_type = TimeSeriesDataType.objects.get(
            category=TimeSeriesDataType.activities, resource='steps')
        return utils.get_fitbit_data(
            self.fbuser, resource_type, base_date=self.base_date,
            period=self.period, end_date=self.end_date)

    def _error_test(self, error):
        with self.assertRaises(error):
            self._mock_time_series(error=error)

    def test_value_error(self):
        """ValueError from the Fitbit.time_series should propagate."""
        self._error_test(ValueError)

    def test_type_error(self):
        """TypeError from the Fitbit.time_series should propagate."""
        self._error_test(TypeError)

    def test_unauthorized(self):
        """HTTPUnauthorized from the Fitbit.time_series should propagate."""
        self._error_test(fitbit_exceptions.HTTPUnauthorized)

    def test_forbidden(self):
        """HTTPForbidden from the Fitbit.time_series should propagate."""
        self._error_test(fitbit_exceptions.HTTPForbidden)

    def test_not_found(self):
        """HTTPNotFound from the Fitbit.time_series should propagate."""
        self._error_test(fitbit_exceptions.HTTPNotFound)

    def test_conflict(self):
        """HTTPConflict from the Fitbit.time_series should propagate."""
        self._error_test(fitbit_exceptions.HTTPConflict)

    def test_server_error(self):
        """HTTPServerError from the Fitbit.time_series should propagate."""
        self._error_test(fitbit_exceptions.HTTPServerError)

    def test_bad_request(self):
        """HTTPBadRequest from the Fitbit.time_series should propagate."""
        self._error_test(fitbit_exceptions.HTTPBadRequest)

    def test_too_many_requests(self):
        """HTTPTooManyRequests from the Fitbit.time_series should propagate."""
        try:
            self._mock_time_series(error=fitbit_exceptions.HTTPTooManyRequests,
                                   error_attrs={'retry_after_secs': 35})
        except fitbit_exceptions.HTTPTooManyRequests:
            self.assertEqual(sys.exc_info()[1].retry_after_secs, 35)
        else:
            assert False, 'Should have thrown exception'

    def test_retrieval(self):
        """get_fitbit_data should return a list of daily steps data."""
        response = {'activities-steps': [1, 2, 3]}
        steps = self._mock_time_series(response=response)
        self.assertEqual(steps, response['activities-steps'])


class TestRetrievalTask(FitappTestBase):
    def setUp(self):
        super(TestRetrievalTask, self).setUp()
        self.category = 'activities'
        self.date = '2013-05-02'
        self.value = 10

    def _receive_fitbit_updates(self, file=False):
        d = json.dumps([{
            'subscriptionId': self.fbuser.user.id,
            'ownerId': self.fbuser.fitbit_user,
            'collectionType': self.category,
            'date': self.date
        }]).encode('utf8')
        kwargs = {'data': d, 'content_type': 'application/json'}
        if file:
            updates_stringio = BytesIO(d)
            updates = InMemoryUploadedFile(
                updates_stringio, None, 'updates', 'text', len(d), None)
            kwargs = {'data': {'updates': updates}}
        res = self.client.post(reverse('fitbit-update'), **kwargs)
        assert res.status_code, 204

    @patch('fitapp.utils.get_fitbit_data')
    def test_subscription_update(self, get_fitbit_data):
        # Check that celery tasks get made when a notification is received
        # from Fitbit.
        get_fitbit_data.return_value = [{'value': self.value}]
        category = getattr(TimeSeriesDataType, self.category)
        resources = TimeSeriesDataType.objects.filter(category=category)
        self._receive_fitbit_updates()
        self.assertEqual(get_fitbit_data.call_count, resources.count())
        # Check that the cache locks have been deleted
        for resource in resources:
            self.assertEqual(
                cache.get('fitapp.get_time_series_data-lock-%s-%s-%s' % (
                    category, resource.resource, self.date)
                ), None)
        date = parser.parse(self.date)
        for tsd in TimeSeriesData.objects.filter(user=self.user, date=date):
            assert tsd.value, self.value

    @patch('fitapp.utils.get_fitbit_data')
    def test_subscription_update_file(self, get_fitbit_data):
        # Check that celery tasks get made when an updates file notification
        # is received from Fitbit.
        get_fitbit_data.return_value = [{'value': self.value}]
        category = getattr(TimeSeriesDataType, self.category)
        resources = TimeSeriesDataType.objects.filter(category=category)
        self._receive_fitbit_updates(file=True)
        self.assertEqual(get_fitbit_data.call_count, resources.count())
        # Check that the cache locks have been deleted
        for resource in resources:
            self.assertEqual(
                cache.get('fitapp.get_time_series_data-lock-%s-%s-%s' % (
                    category, resource.resource, self.date)
                ), None)
        date = parser.parse(self.date)
        for tsd in TimeSeriesData.objects.filter(user=self.user, date=date):
            assert tsd.value, self.value

    @patch('fitapp.utils.get_fitbit_data')
    @patch('django.core.cache.cache.add')
    def test_subscription_update_locked(self, mock_add, get_fitbit_data):
        # Check that celery tasks do not get made when a notification is
        # received from Fitbit, but there is already a matching task in
        # progress
        mock_add.return_value = False
        self.assertEqual(TimeSeriesData.objects.count(), 0)
        self._receive_fitbit_updates()
        self.assertEqual(get_fitbit_data.call_count, 0)
        self.assertEqual(TimeSeriesData.objects.count(), 0)

    @patch('fitapp.utils.get_fitbit_data')
    @patch('fitapp.tasks.get_time_series_data.retry')
    def test_subscription_update_too_many(self, mock_retry, get_fitbit_data):
        # Check that celery tasks get postponed if the rate limit is hit
        mock_retry.return_value = celery.exceptions.Retry()
        exc = fitbit_exceptions.HTTPTooManyRequests(self._error_response())
        exc.retry_after_secs = 21
        get_fitbit_data.side_effect = exc
        category = getattr(TimeSeriesDataType, self.category)
        resources = TimeSeriesDataType.objects.filter(category=category)
        self.assertEqual(TimeSeriesData.objects.count(), 0)
        self._receive_fitbit_updates()
        self.assertEqual(get_fitbit_data.call_count, resources.count())
        self.assertEqual(TimeSeriesData.objects.count(), 0)
        mock_retry.assert_called_with(exc, countdown=21)

    def test_problem_queueing_task(self):
        get_time_series_data = MagicMock()
        # If queueing the task raises an exception, it doesn't propagate
        get_time_series_data.apply_async.side_effect = Exception
        try:
            self._receive_fitbit_updates()
        except:
            assert False, 'Any errors should be captured in the view'


class RetrievalViewTestBase(object):
    """Base methods for the get_steps view."""
    url_name = 'fitbit-steps'
    valid_periods = utils.get_valid_periods()

    def setUp(self):
        super(RetrievalViewTestBase, self).setUp()
        self.period = '30d'
        self.base_date = '2012-06-06'
        self.end_date = '2012-07-07'

    def _check_response(self, response, code, objects=None, error_msg=None):
        objects = objects or []
        self.assertEqual(response.status_code, 200)
        data = json.loads(response.content.decode('utf8'))
        self.assertEqual(data['meta']['status_code'], code, error_msg)
        self.assertEqual(data['meta']['total_count'], len(objects), error_msg)
        self.assertEqual(data['objects'], objects, error_msg)

    def test_not_authenticated(self):
        """Status code should be 101 when user isn't logged in."""
        self.client.logout()
        response = self._get(get_kwargs=self._data())
        self._check_response(response, 101)
        self.assertEqual(UserFitbit.objects.count(), 1)

    def test_not_active(self):
        """Status code should be 101 when user isn't active."""
        self.user.is_active = False
        self.user.save()
        response = self._get(get_kwargs=self._data())
        self._check_response(response, 101)
        self.assertEqual(UserFitbit.objects.count(), 1)

    @override_settings(FITAPP_SUBSCRIBE=False)
    def test_not_integrated(self):
        """
        Status code should be 102 when an unsubscribed user is not integrated.
        """
        self.fbuser.delete()
        response = self._get(get_kwargs=self._data())
        self._check_response(response, 102)
        self.assertEqual(UserFitbit.objects.count(), 0)

    @override_settings(FITAPP_SUBSCRIBE=False)
    def test_invalid_credentials_unauthorized(self):
        """
        Status code should be 103 & credentials should be deleted when user
        integration is invalid.
        """
        response = self._mock_utility(get_kwargs=self._data(),
                                      error=fitbit_exceptions.HTTPUnauthorized)
        self._check_response(response, 103)
        self.assertEqual(UserFitbit.objects.count(), 0)

    @override_settings(FITAPP_SUBSCRIBE=False)
    def test_invalid_credentials_forbidden(self):
        """
        Status code should be 103 & credentials should be deleted when user
        integration is invalid.
        """
        response = self._mock_utility(get_kwargs=self._data(),
                                      error=fitbit_exceptions.HTTPForbidden)
        self._check_response(response, 103)
        self.assertEqual(UserFitbit.objects.count(), 0)

    @override_settings(FITAPP_SUBSCRIBE=False)
    def test_rate_limited(self):
        """Status code should be 105 when Fitbit rate limit is hit."""
        response = self._mock_utility(get_kwargs=self._data(),
                                      error=fitbit_exceptions.HTTPConflict)
        self._check_response(response, 105)

    @override_settings(FITAPP_SUBSCRIBE=False)
    def test_fitbit_error(self):
        """Status code should be 106 when Fitbit server error occurs."""
        response = self._mock_utility(get_kwargs=self._data(),
                                      error=fitbit_exceptions.HTTPServerError)
        self._check_response(response, 106)

    def test_405(self):
        """View should not respond to anything but a GET request."""
        url = reverse('fitbit-data', args=['activities', 'steps'])
        for method in (self.client.post, self.client.head,
                       self.client.options, self.client.put,
                       self.client.delete):
            response = method(url)
            self.assertEqual(response.status_code, 405)

    def test_ambiguous(self):
        """Status code should be 104 when both period & end_date are given."""
        data = {'end_date': self.end_date, 'period': self.period,
                'base_date': self.base_date}
        response = self._get(get_kwargs=data)
        self._check_response(response, 104)


class TestRetrievePeriod(RetrievalViewTestBase, FitappTestBase):

    def _data(self):
        return {'base_date': self.base_date, 'period': self.period}

    def test_no_period(self):
        """Status code should be 104 when no period is given."""
        data = self._data()
        data.pop('period')
        response = self._get(get_kwargs=data)
        self._check_response(response, 104)

    def test_bad_period(self):
        """Status code should be 104 when invalid period is given."""
        self.period = 'bad'
        response = self._get(get_kwargs=self._data())
        self._check_response(response, 104)

    @freeze_time('2012-06-06')
    def test_no_base_date(self):
        """Base date should be optional for period request."""
        data = self._data()
        data.pop('base_date')
        steps = [{'dateTime': '2012-06-07', 'value': '10'}]
        TimeSeriesData.objects.create(
            user=self.user,
            resource_type=TimeSeriesDataType.objects.get(
                category=TimeSeriesDataType.activities, resource='steps'),
            date=steps[0]['dateTime'],
            value=steps[0]['value']
        )
        response = self._mock_utility(response=steps, get_kwargs=data)
        self._check_response(response, 100, steps)

    def test_bad_base_date(self):
        """Status code should be 104 when invalid base date is given."""
        self.base_date = 'bad'
        response = self._get(get_kwargs=self._data())
        self._check_response(response, 104)

    def test_period(self):
        steps = [{'dateTime': '2012-06-07', 'value': '10'}]
        TimeSeriesData.objects.create(
            user=self.user,
            resource_type=TimeSeriesDataType.objects.get(
                category=TimeSeriesDataType.activities, resource='steps'),
            date=steps[0]['dateTime'],
            value=steps[0]['value']
        )
        for period in self.valid_periods:
            self.period = period
            data = self._data()
            response = self._mock_utility(response=steps, get_kwargs=data)
            error_msg = 'Should be able to retrieve data for {0}.'.format(
                self.period)
            self._check_response(response, 100, steps, error_msg)

    def test_period_not_integrated(self):
        """
        Period data is returned to a subscribed user who is not integrated
        """
        self.fbuser.delete()
        steps = [{'dateTime': '2012-06-07', 'value': '10'}]
        TimeSeriesData.objects.create(
            user=self.user,
            resource_type=TimeSeriesDataType.objects.get(
                category=TimeSeriesDataType.activities, resource='steps'),
            date=steps[0]['dateTime'],
            value=steps[0]['value']
        )
        for period in self.valid_periods:
            self.period = period
            data = self._data()
            response = self._mock_utility(response=steps, get_kwargs=data)
            error_msg = 'Should be able to retrieve data for {0}.'.format(
                self.period)
            self._check_response(response, 100, steps, error_msg)


class TestRetrieveRange(RetrievalViewTestBase, FitappTestBase):

    def _data(self):
        return {'base_date': self.base_date, 'end_date': self.end_date}

    def test_range__no_base_date(self):
        """Status code should be 104 when no base date is given."""
        data = self._data()
        data.pop('base_date')
        response = self._get(get_kwargs=data)
        self._check_response(response, 104)

    def test_range__bad_base_date(self):
        """Status code should be 104 when invalid base date is given."""
        self.base_date = 'bad'
        response = self._get(get_kwargs=self._data())
        self._check_response(response, 104)

    def test_range__no_end_date(self):
        """Status code should be 104 when no end date is given."""
        data = self._data()
        data.pop('end_date')
        response = self._get(get_kwargs=data)
        self._check_response(response, 104)

    def test_range__bad_end_date(self):
        """Status code should be 104 when invalid end date is given."""
        self.end_date = 'bad'
        response = self._get(get_kwargs=self._data())
        self._check_response(response, 104)

    def test_range(self):
        steps = [{'dateTime': '2012-06-07', 'value': '10'}]
        TimeSeriesData.objects.create(
            user=self.user,
            resource_type=TimeSeriesDataType.objects.get(
                category=TimeSeriesDataType.activities, resource='steps'),
            date=steps[0]['dateTime'],
            value=steps[0]['value']
        )

        response = self._mock_utility(response=steps,
                                      get_kwargs=self._data())
        self._check_response(response, 100, steps)

    def test_range_not_integrated(self):
        """
        Range data is returned to a subscribed user who is not integrated
        """
        self.fbuser.delete()
        steps = [{'dateTime': '2012-06-07', 'value': '10'}]
        TimeSeriesData.objects.create(
            user=self.user,
            resource_type=TimeSeriesDataType.objects.get(
                category=TimeSeriesDataType.activities, resource='steps'),
            date=steps[0]['dateTime'],
            value=steps[0]['value']
        )

        response = self._mock_utility(response=steps,
                                      get_kwargs=self._data())
        self._check_response(response, 100, steps)



from django.conf.urls import patterns, url

from . import views


urlpatterns = patterns('',

    # OAuth authentication
    url(r'^login/$',
            views.login, name='fitbit-login'),
    url(r'^complete/$',
            views.complete, name='fitbit-complete'),
    url(r'^error/$',
            views.error, name='fitbit-error'),
    url(r'^logout/$',
            views.logout, name='fitbit-logout'),

    # Subscriber callback for near realtime updates
    url(r'^update/$',
            views.update, name='fitbit-update'),

    # Fitbit data retrieval
    url(r'^get_data/(?P<category>[\w]+)/(?P<resource>[/\w]+)/$',
            views.get_data, name='fitbit-data'),
    url(r'^get_steps/$', views.get_steps, name='fitbit-steps')
)



# Your Fitbit access credentials, which must be requested from Fitbit.
# You must provide these in your project's settings.
FITAPP_CONSUMER_KEY = None
FITAPP_CONSUMER_SECRET = None

# Where to redirect to after Fitbit authentication is successfully completed.
FITAPP_LOGIN_REDIRECT = '/'

# Where to redirect to after Fitbit authentication credentials have been
# removed.
FITAPP_LOGOUT_REDIRECT = '/'

# By default, don't subscribe to user data. Set this to true to subscribe.
FITAPP_SUBSCRIBE = False

# The template to use when an unavoidable error occurs during Fitbit
# integration.
FITAPP_ERROR_TEMPLATE = 'fitapp/error.html'

# The default message used by the fitbit_integration_warning decorator to
# inform the user about Fitbit integration. If a callable is given, it is
# called with the request as the only parameter to get the final value for the
# message.
FITAPP_DECORATOR_MESSAGE = 'This page requires Fitbit integration.'



from mock import patch, Mock
import random
try:
    from urllib.parse import urlencode
    from string import ascii_letters
except:
    # Python 2.x
    from urllib import urlencode
    from string import letters as ascii_letters

from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.test import TestCase

from fitbit.api import Fitbit

from fitapp.models import UserFitbit


class MockClient(object):

    def __init__(self, **kwargs):
        self.user_id = kwargs.get('user_id', None)
        self.resource_owner_key = kwargs.get('resource_owner_key', None)
        self.resource_owner_secret = kwargs.get('resource_owner_secret', None)
        self.error = kwargs.get('error', None)

    def authorize_token_url(self, *args, **kwargs):
        return '/test'

    def fetch_request_token(self, *args, **kwargs):
        return None

    def fetch_access_token(self, *args, **kwargs):
        if self.error:
            raise self.error('')
        return {
            'resource_owner_key': self.resource_owner_key,
            'resource_owner_secret': self.resource_owner_secret
        }

    def make_request(self, *args, **kwargs):
        response = Mock()
        response.status_code = 204
        response.content = "{}".encode('utf8')
        return response


class FitappTestBase(TestCase):
    TEST_SERVER = 'http://testserver'

    def setUp(self):
        self.username = self.random_string(25)
        self.password = self.random_string(25)
        self.user = self.create_user(username=self.username,
                                     password=self.password)
        self.fbuser = self.create_userfitbit(user=self.user)

        self.client.login(username=self.username, password=self.password)

    def random_string(self, length=255, extra_chars=''):
        chars = ascii_letters + extra_chars
        return ''.join([random.choice(chars) for i in range(length)])

    def create_user(self, username=None, email=None, password=None, **kwargs):
        username = username or self.random_string(25)
        email = email or '{0}@{1}.com'.format(self.random_string(25),
                                              self.random_string(10))
        password = password or self.random_string(25)
        user = User.objects.create_user(username, email, password)
        User.objects.filter(pk=user.pk).update(**kwargs)
        user = User.objects.get(pk=user.pk)
        return user

    def create_userfitbit(self, **kwargs):
        defaults = {
            'user': kwargs.pop('user', self.create_user()),
            'fitbit_user': kwargs.pop('fitbit_user', self.random_string(25)),
            'auth_token': self.random_string(25),
            'auth_secret': self.random_string(25),
        }
        defaults.update(kwargs)
        return UserFitbit.objects.create(**defaults)

    def create_fitbit(self, **kwargs):
        defaults = {
            'client_key': self.random_string(25),
            'client_secret': self.random_string(25),
        }
        defaults.update(kwargs)
        return Fitbit(**defaults)

    def assertRedirectsNoFollow(self, response, url, status_code=302):
        """
        Workaround to test whether a response redirects to another URL without
        loading the page at that URL.
        """
        self.assertEqual(response.status_code, status_code)
        full_url = self.TEST_SERVER + url
        self.assertEqual(response._headers['location'][1], full_url)

    def _get(self, url_name=None, url_kwargs=None, get_kwargs=None, **kwargs):
        """Convenience wrapper for test client GET request."""
        url_name = url_name or self.url_name
        url = reverse(url_name, kwargs=url_kwargs)  # Base URL.

        # Add GET parameters.
        if get_kwargs:
            url += '?' + urlencode(get_kwargs)

        return self.client.get(url, **kwargs)

    def _set_session_vars(self, **kwargs):
        session = self.client.session
        for key, value in kwargs.items():
            session[key] = value
        try:
            session.save()  # Only available on authenticated sessions.
        except AttributeError:
            pass

    def _error_response(self):
        error_response = Mock(['content'])
        error_response.content = '{"errors": []}'.encode('utf8')
        return error_response

    @patch('fitbit.api.FitbitOauthClient')
    def _mock_client(self, client=None, client_kwargs=None, **kwargs):
        client_kwargs = client_kwargs or {}
        client.return_value = MockClient(**client_kwargs)
        return self._get(**kwargs)

    @patch('fitapp.utils.get_fitbit_data')
    def _mock_utility(self, utility=None, error=None, response=None, **kwargs):
        if error:
            utility.side_effect = error(self._error_response())
        elif response:
            utility.return_value = response
        return self._get(**kwargs)



from functools import wraps

from django.contrib import messages

from . import utils


def fitbit_integration_warning(msg=None):
    """
    Adds a message to inform the user about Fitbit integration if their
    account is not already integrated with Fitbit.

    :param msg: The message text to use if the user is not integrated. If msg
        is a callable, it is called with the request as the only parameter.
        Otherwise, msg is passed in as the message text. If no message is
        provided, the value in :ref:`FITAPP_DECORATOR_MESSAGE` is used.

    This decorator does not prevent the user from seeing the view if they are
    not integrated with Fitbit - it only adds a message to the user's
    messages. If you would like to change the behavior of your view based on
    whether the user is integrated, you can check the user's status by using
    :py:func:`fitapp.utils.is_integrated`.

    Example::

        from django.http import HttpResponse
        from django.contrib.auth.decorators import login_required
        from fitapp.decorators import fitbit_integration_warning

        @fitbit_integration_warning(msg="Integrate your account with Fitbit!")
        @login_required
        def my_view(request):
            return HttpResponse('Visible to authenticated users regardless' +
                    'of Fitbit integration status')

    In this example, the ``fitbit_integration_warning`` decorator only
    operates if the user is logged in. The view content is visible to all
    users who are logged in, regardless of Fitbit integration status.

    The template(s) for any view with this decorator should display a user's
    messages.
    """
    if not msg:
        msg = utils.get_setting('FITAPP_DECORATOR_MESSAGE')
    def inner_decorator(view_func):
        def wrapped(request, *args, **kwargs):
            user = request.user
            if not utils.is_integrated(user):
                text = msg(request) if callable(msg) else msg
                messages.error(request, text)
            return view_func(request, *args, **kwargs)
        return wraps(view_func)(wrapped)
    return inner_decorator



from django import template

from fitapp import utils


register= template.Library()


@register.filter
def is_integrated_with_fitbit(user):
    """Returns ``True`` if we have Oauth info for the user.

    For example::

        {% if request.user|is_integrated_with_fitbit %}
            do something
        {% else %}
            do something else
        {% endif %}
    """
    return utils.is_integrated(user)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='TimeSeriesData',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('date', models.DateField()),
                ('value', models.CharField(default=None, max_length=32, null=True)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='TimeSeriesDataType',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('category', models.IntegerField(choices=[(0, b'foods'), (1, b'activities'), (2, b'sleep'), (3, b'body')])),
                ('resource', models.CharField(max_length=128)),
            ],
            options={
                'ordering': ['category', 'resource'],
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name='UserFitbit',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('fitbit_user', models.CharField(unique=True, max_length=32)),
                ('auth_token', models.TextField()),
                ('auth_secret', models.TextField()),
                ('user', models.OneToOneField(to=settings.AUTH_USER_MODEL)),
            ],
            options={
            },
            bases=(models.Model,),
        ),
        migrations.AlterUniqueTogether(
            name='timeseriesdatatype',
            unique_together=set([('category', 'resource')]),
        ),
        migrations.AddField(
            model_name='timeseriesdata',
            name='resource_type',
            field=models.ForeignKey(to='fitapp.TimeSeriesDataType'),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name='timeseriesdata',
            name='user',
            field=models.ForeignKey(to=settings.AUTH_USER_MODEL),
            preserve_default=True,
        ),
        migrations.AlterUniqueTogether(
            name='timeseriesdata',
            unique_together=set([('user', 'resource_type', 'date')]),
        ),
    ]



from __future__ import absolute_import

import os

from celery import Celery
from django.conf import settings


# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'fitapp.defaults')

app = Celery('fitapp_tests')
app.conf['CELERY_ALWAYS_EAGER'] = True
app.autodiscover_tasks(lambda: settings.INSTALLED_APPS, force=True)



from __future__ import absolute_import

from fitapp.tests.test_retrieval import *
from fitapp.tests.test_integration import *

# This will make sure the app is always imported when
# Django starts so that shared_task will use this app.
from .celery import app



from django import forms

from . import utils


INPUT_FORMATS = ['%Y-%m-%d']


class PeriodForm(forms.Form):
    """Data necessary to request Fitbit data from a period of time."""
    base_date = forms.DateField(input_formats=INPUT_FORMATS, required=False)
    period = forms.ChoiceField(choices=[])

    def __init__(self, *args, **kwargs):
        super(PeriodForm, self).__init__(*args, **kwargs)
        PERIOD_CHOICES = [(p, p) for p in utils.get_valid_periods()]
        self.fields['period'].choices = PERIOD_CHOICES

    def get_fitbit_data(self):
        if self.is_valid():
            return {
                'base_date': self.cleaned_data['base_date'] or 'today',
                'period': self.cleaned_data['period'],
            }


class RangeForm(forms.Form):
    """Data necessary to request Fitbit data from a specific time range."""
    base_date = forms.DateField(input_formats=INPUT_FORMATS)
    end_date = forms.DateField(input_formats=INPUT_FORMATS)

    def get_fitbit_data(self):
        if self.is_valid():
            return {
                'base_date': self.cleaned_data['base_date'],
                'end_date': self.cleaned_data['end_date'],
            }



from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from fitbit import Fitbit

from fitapp.utils import create_fitbit, get_setting


class TestFitappUtilities(TestCase):
    def test_create_fitbit(self):
        """
        Check that the create_fitbit utility creates a Fitbit object and
        produces an error is raised when the consumer key or secret aren't set.
        """
        with self.settings(FITAPP_CONSUMER_KEY=None,
                           FITAPP_CONSUMER_SECRET=None):
            self.assertRaises(ImproperlyConfigured, create_fitbit)
        with self.settings(FITAPP_CONSUMER_KEY='',
                           FITAPP_CONSUMER_SECRET=None):
            self.assertRaises(ImproperlyConfigured, create_fitbit)
        with self.settings(FITAPP_CONSUMER_KEY=None,
                           FITAPP_CONSUMER_SECRET=''):
            self.assertRaises(ImproperlyConfigured, create_fitbit)
        self.assertEqual(type(create_fitbit()), Fitbit)

    def test_get_setting_error(self):
        """
        Check that an error is raised when trying to get a nonexistent setting.
        """
        self.assertRaises(ImproperlyConfigured, get_setting, 'DOES_NOT_EXIST')



from django.conf import settings
from django.contrib import messages
from django.contrib.auth.models import AnonymousUser
from django.core.urlresolvers import reverse
from django.http import HttpRequest
from fitbit.exceptions import HTTPConflict
from mock import MagicMock, patch

from fitapp import utils
from fitapp.decorators import fitbit_integration_warning
from fitapp.models import UserFitbit, TimeSeriesDataType
from fitapp.tasks import subscribe, unsubscribe

from .base import FitappTestBase


class TestIntegrationUtility(FitappTestBase):

    def test_is_integrated(self):
        """Users with stored OAuth information are integrated."""
        self.assertTrue(utils.is_integrated(self.user))

    def test_is_not_integrated(self):
        """User is not integrated if we have no OAuth data for them."""
        UserFitbit.objects.all().delete()
        self.assertFalse(utils.is_integrated(self.user))

    def test_unauthenticated(self):
        """User is not integrated if they aren't logged in."""
        user = AnonymousUser()
        self.assertFalse(utils.is_integrated(user))


class TestIntegrationDecorator(FitappTestBase):

    def setUp(self):
        super(TestIntegrationDecorator, self).setUp()
        self.fake_request = HttpRequest()
        self.fake_request.user = self.user
        self.fake_view = lambda request: "hello"
        self.messages = []

    def _mock_decorator(self, msg=None):
        def mock_error(request, message, *args, **kwargs):
            self.messages.append(message)

        with patch.object(messages, 'error', mock_error) as error:
            return fitbit_integration_warning(msg=msg)(self.fake_view)(
                    self.fake_request)

    def test_unauthenticated(self):
        """Message should be added if user is not logged in."""
        self.fake_request.user = AnonymousUser()
        results = self._mock_decorator()

        self.assertEqual(results, "hello")
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0],
                utils.get_setting('FITAPP_DECORATOR_MESSAGE'))

    def test_is_integrated(self):
        """Decorator should have no effect if user is integrated."""
        results = self._mock_decorator()

        self.assertEqual(results, "hello")
        self.assertEqual(len(self.messages), 0)

    def test_is_not_integrated(self):
        """Message should be added if user is not integrated."""
        UserFitbit.objects.all().delete()
        results = self._mock_decorator()

        self.assertEqual(results, "hello")
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0],
                utils.get_setting('FITAPP_DECORATOR_MESSAGE'))

    def test_custom_msg(self):
        """Decorator should support a custom message string."""
        UserFitbit.objects.all().delete()
        msg = "customized"
        results = self._mock_decorator(msg)

        self.assertEqual(results, "hello")
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0], "customized")

    def test_custom_msg_func(self):
        """Decorator should support a custom message function."""
        UserFitbit.objects.all().delete()
        msg = lambda request: "message to {0}".format(request.user)
        results = self._mock_decorator(msg)

        self.assertEqual(results, "hello")
        self.assertEqual(len(self.messages), 1)
        self.assertEqual(self.messages[0], msg(self.fake_request))


class TestLoginView(FitappTestBase):
    url_name = 'fitbit-login'

    def test_get(self):
        """
        Login view should generate & store a request token then
        redirect to an authorization URL.
        """
        response = self._mock_client()
        self.assertRedirectsNoFollow(response, '/test')
        self.assertTrue('token' in self.client.session)
        self.assertEqual(UserFitbit.objects.count(), 1)

    def test_unauthenticated(self):
        """User must be logged in to access Login view."""
        self.client.logout()
        response = self._get()
        self.assertEqual(response.status_code, 302)
        self.assertEqual(UserFitbit.objects.count(), 1)

    def test_unintegrated(self):
        """Fitbit credentials not required to access Login view."""
        self.fbuser.delete()
        response = self._mock_client()
        self.assertRedirectsNoFollow(response, '/test')
        self.assertTrue('token' in self.client.session)
        self.assertEqual(UserFitbit.objects.count(), 0)

    def test_next(self):
        response = self._mock_client(get_kwargs={'next': '/next'})
        self.assertRedirectsNoFollow(response, '/test')
        self.assertEqual(self.client.session.get('fitbit_next', None),
                '/next')
        self.assertEqual(UserFitbit.objects.count(), 1)


class TestCompleteView(FitappTestBase):
    url_name = 'fitbit-complete'
    resource_owner_key = 'abc'
    resource_owner_secret = '123'
    user_id = 'userid'

    def setUp(self):
        super(TestCompleteView, self).setUp()
        self.fbuser.delete()

    def _get(self, use_token=True, use_verifier=True, **kwargs):
        if use_token:
            self._set_session_vars(token='token')
        get_kwargs = kwargs.pop('get_kwargs', {})
        if use_verifier:
            get_kwargs.update({'oauth_verifier': 'hello'})
        return super(TestCompleteView, self)._get(get_kwargs=get_kwargs,
                **kwargs)

    def _mock_client(self, client_kwargs=None, **kwargs):
        client_kwargs = client_kwargs or {}
        defaults = {
            'resource_owner_key': self.resource_owner_key,
            'resource_owner_secret': self.resource_owner_secret,
            'user_id': self.user_id,
        }
        defaults.update(client_kwargs)
        return super(TestCompleteView, self)._mock_client(
                client_kwargs=defaults, **kwargs)

    @patch('fitapp.tasks.subscribe.apply_async')
    @patch('fitapp.tasks.get_time_series_data.apply_async')
    def test_get(self, tsd_apply_async, sub_apply_async):
        """Complete view should fetch & store user's access credentials."""
        response = self._mock_client()
        self.assertRedirectsNoFollow(
             response, utils.get_setting('FITAPP_LOGIN_REDIRECT'))
        fbuser = UserFitbit.objects.get()
        sub_apply_async.assert_called_once_with(
            (fbuser.fitbit_user, settings.FITAPP_SUBSCRIBER_ID), countdown=5)
        tsdts = TimeSeriesDataType.objects.all()
        self.assertEqual(tsd_apply_async.call_count, tsdts.count())
        for i, _type in enumerate(tsdts):
            tsd_apply_async.assert_any_call(
                (fbuser.fitbit_user, _type.category, _type.resource,),
                countdown=10 + (i * 5))
        self.assertEqual(fbuser.user, self.user)
        self.assertEqual(fbuser.auth_token, self.resource_owner_key)
        self.assertEqual(fbuser.auth_secret, self.resource_owner_secret)
        self.assertEqual(fbuser.fitbit_user, self.user_id)

    @patch('fitapp.tasks.subscribe.apply_async')
    @patch('fitapp.tasks.get_time_series_data.apply_async')
    def test_get(self, tsd_apply_async, sub_apply_async):
        """
        Complete view redirect to the error view if a user attempts to connect
        an already integrated fitbit user to a second user.
        """
        self.create_userfitbit(user=self.user, fitbit_user=self.user_id)
        username2 = '%s2' % self.username
        user2 = self.create_user(username=username2, password=self.password)
        self.client.logout()
        self.client.login(username=username2, password=self.password)
        response = self._mock_client()
        self.assertRedirectsNoFollow(response, reverse('fitbit-error'))
        self.assertEqual(UserFitbit.objects.all().count(), 1)
        self.assertEqual(sub_apply_async.call_count, 0)
        self.assertEqual(tsd_apply_async.call_count, 0)

    def test_unauthenticated(self):
        """User must be logged in to access Complete view."""
        self.client.logout()
        response = self._mock_client()
        self.assertEqual(response.status_code, 302)
        self.assertEqual(UserFitbit.objects.count(), 0)

    @patch('fitapp.tasks.subscribe.apply_async')
    @patch('fitapp.tasks.get_time_series_data.apply_async')
    def test_next(self, tsd_apply_async, sub_apply_async):
        """
        Complete view should redirect to session['fitbit_next'] if available.
        """
        self._set_session_vars(fitbit_next='/test')
        response = self._mock_client()
        self.assertRedirectsNoFollow(response, '/test')
        fbuser = UserFitbit.objects.get()
        sub_apply_async.assert_called_once_with(
            (fbuser.fitbit_user, settings.FITAPP_SUBSCRIBER_ID), countdown=5)
        self.assertEqual(tsd_apply_async.call_count,
                         TimeSeriesDataType.objects.count())
        self.assertEqual(fbuser.user, self.user)
        self.assertEqual(fbuser.auth_token, self.resource_owner_key)
        self.assertEqual(fbuser.auth_secret, self.resource_owner_secret)
        self.assertEqual(fbuser.fitbit_user, self.user_id)

    def test_access_error(self):
        """
        Complete view should redirect to error if access token is
        inaccessible.
        """
        response = self._mock_client(client_kwargs={'error': Exception})
        self.assertRedirectsNoFollow(response, reverse('fitbit-error'))
        self.assertEqual(UserFitbit.objects.count(), 0)

    def test_no_token(self):
        """Complete view should redirect to error if token isn't in session."""
        response = self._get(use_token=False)
        self.assertRedirectsNoFollow(response, reverse('fitbit-error'))
        self.assertEqual(UserFitbit.objects.count(), 0)

    def test_no_verifier(self):
        """
        Complete view should redirect to error if verifier param is not
        present.
        """
        response = self._get(use_verifier=False)
        self.assertRedirectsNoFollow(response, reverse('fitbit-error'))
        self.assertEqual(UserFitbit.objects.count(), 0)

    @patch('fitapp.tasks.subscribe.apply_async')
    @patch('fitapp.tasks.get_time_series_data.apply_async')
    def test_integrated(self, tsd_apply_async, sub_apply_async):
        """
        Complete view should overwrite existing credentials for this user.
        """
        self.fbuser = self.create_userfitbit(user=self.user)
        response = self._mock_client()
        fbuser = UserFitbit.objects.get()
        sub_apply_async.assert_called_with(
            (fbuser.fitbit_user, settings.FITAPP_SUBSCRIBER_ID), countdown=5)
        self.assertEqual(tsd_apply_async.call_count,
                         TimeSeriesDataType.objects.count())
        self.assertEqual(fbuser.user, self.user)
        self.assertEqual(fbuser.auth_token, self.resource_owner_key)
        self.assertEqual(fbuser.auth_secret, self.resource_owner_secret)
        self.assertEqual(fbuser.fitbit_user, self.user_id)
        self.assertRedirectsNoFollow(
            response, utils.get_setting('FITAPP_LOGIN_REDIRECT'))


class TestErrorView(FitappTestBase):
    url_name = 'fitbit-error'

    def test_get(self):
        """Should be able to retrieve Error page."""
        response = self._get()
        self.assertEqual(response.status_code, 200)

    def test_unauthenticated(self):
        """User must be logged in to access Error view."""
        self.client.logout()
        response = self._get()
        self.assertEqual(response.status_code, 302)

    def test_unintegrated(self):
        """No Fitbit credentials required to access Error view."""
        self.fbuser.delete()
        response = self._get()
        self.assertEqual(response.status_code, 200)


class TestLogoutView(FitappTestBase):
    url_name = 'fitbit-logout'

    @patch('fitapp.tasks.unsubscribe.apply_async')
    def test_get(self, apply_async):
        """Logout view should remove associated UserFitbit and redirect."""
        response = self._get()
        apply_async.assert_called_once_with(kwargs=self.fbuser.get_user_data(),
                                            countdown=5)
        self.assertRedirectsNoFollow(response,
                utils.get_setting('FITAPP_LOGIN_REDIRECT'))
        self.assertEqual(UserFitbit.objects.count(), 0)

    def test_unauthenticated(self):
        """User must be logged in to access Logout view."""
        self.client.logout()
        response = self._get()
        self.assertEqual(response.status_code, 302)
        self.assertEqual(UserFitbit.objects.count(), 1)

    def test_unintegrated(self):
        """No Fitbit credentials required to access Logout view."""
        self.fbuser.delete()
        response = self._get()
        self.assertRedirectsNoFollow(response,
                utils.get_setting('FITAPP_LOGIN_REDIRECT'))
        self.assertEqual(UserFitbit.objects.count(), 0)

    @patch('fitapp.tasks.unsubscribe.apply_async')
    def test_next(self, apply_async):
        """Logout view should redirect to GET['next'] if available."""
        response = self._get(get_kwargs={'next': '/test'})
        apply_async.assert_called_with(kwargs=self.fbuser.get_user_data(),
                                       countdown=5)
        self.assertRedirectsNoFollow(response, '/test')
        self.assertEqual(UserFitbit.objects.count(), 0)


class TestSubscription(FitappTestBase):
    @patch('fitbit.Fitbit.subscription')
    def test_subscribe(self, subscription):
        subscribe.apply_async((self.fbuser.fitbit_user, 1,))
        subscription.assert_called_once_with(self.user.id, 1,)

    @patch('fitbit.Fitbit.subscription')
    def test_subscribe_error(self, subscription):
        subscription.side_effect = HTTPConflict
        apply_result = subscribe.apply_async((self.fbuser.fitbit_user, 1,))
        self.assertEqual(apply_result.status, 'REJECTED')
        subscription.assert_called_once_with(self.user.id, 1,)

    @patch('fitbit.Fitbit.subscription')
    @patch('fitbit.Fitbit.list_subscriptions')
    def test_unsubscribe(self, list_subscriptions, subscription):
        sub = {
            'ownerId': self.fbuser.fitbit_user,
            'subscriberId': '1',
            'subscriptionId': str(self.user.id).encode('utf8'),
            'collectionType': 'user',
            'ownerType': 'user'
        }
        list_subscriptions.return_value = {'apiSubscriptions': [sub]}
        unsubscribe.apply_async(kwargs=self.fbuser.get_user_data())
        list_subscriptions.assert_called_once_with()
        subscription.assert_called_once_with(
            sub['subscriptionId'], sub['subscriberId'], method="DELETE")

    @patch('fitbit.Fitbit.subscription')
    @patch('fitbit.Fitbit.list_subscriptions')
    def test_unsubscribe_error(self, list_subscriptions, subscription):
        list_subscriptions.side_effect = HTTPConflict
        result = unsubscribe.apply_async(kwargs=self.fbuser.get_user_data())
        self.assertEqual(result.status, 'REJECTED')
        list_subscriptions.assert_called_once_with()
        self.assertEqual(subscription.call_count, 0)



import json

from dateutil import parser
from dateutil.relativedelta import relativedelta
from django.contrib.auth.decorators import login_required
from django.contrib.auth.signals import user_logged_in
from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import reverse
from django.db import IntegrityError
from django.dispatch import receiver
from django.http import HttpResponse, Http404
from django.shortcuts import redirect, render
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_GET, require_POST

from fitbit.exceptions import (HTTPUnauthorized, HTTPForbidden, HTTPConflict,
                               HTTPServerError)

from . import forms
from . import utils
from .models import UserFitbit, TimeSeriesData, TimeSeriesDataType
from .tasks import get_time_series_data, subscribe, unsubscribe

try:
    from types import StringType, UnicodeType
    STRING_TYPES = [StringType, UnicodeType]
except ImportError:  # Python 3
    STRING_TYPES = [str]


@login_required
def login(request):
    """
    Begins the OAuth authentication process by obtaining a Request Token from
    Fitbit and redirecting the user to the Fitbit site for authorization.

    When the user has finished at the Fitbit site, they will be redirected
    to the :py:func:`fitapp.views.complete` view.

    If 'next' is provided in the GET data, it is saved in the session so the
    :py:func:`fitapp.views.complete` view can redirect the user to that URL
    upon successful authentication.

    URL name:
        `fitbit-login`
    """
    next_url = request.GET.get('next', None)
    if next_url:
        request.session['fitbit_next'] = next_url
    else:
        request.session.pop('fitbit_next', None)

    callback_uri = request.build_absolute_uri(reverse('fitbit-complete'))
    fb = utils.create_fitbit(callback_uri=callback_uri)
    token = fb.client.fetch_request_token()
    token_url = fb.client.authorize_token_url()
    request.session['token'] = token
    return redirect(token_url)


@login_required
def complete(request):
    """
    After the user authorizes us, Fitbit sends a callback to this URL to
    complete authentication.

    If there was an error, the user is redirected again to the `error` view.

    If the authorization was successful, the credentials are stored for us to
    use later, and the user is redirected. If 'next_url' is in the request
    session, the user is redirected to that URL. Otherwise, they are
    redirected to the URL specified by the setting
    :ref:`FITAPP_LOGIN_REDIRECT`.

    If :ref:`FITAPP_SUBSCRIBE` is set to True, add a subscription to user
    data at this time.

    URL name:
        `fitbit-complete`
    """
    fb = utils.create_fitbit()
    try:
        token = request.session.pop('token')
        verifier = request.GET.get('oauth_verifier')
    except KeyError:
        return redirect(reverse('fitbit-error'))
    try:
        fb.client.fetch_access_token(verifier, token=token)
    except:
        return redirect(reverse('fitbit-error'))

    if UserFitbit.objects.filter(fitbit_user=fb.client.user_id).exists():
        return redirect(reverse('fitbit-error'))

    fbuser, _ = UserFitbit.objects.get_or_create(user=request.user)
    fbuser.auth_token = fb.client.resource_owner_key
    fbuser.auth_secret = fb.client.resource_owner_secret
    fbuser.fitbit_user = fb.client.user_id
    fbuser.save()

    # Add the Fitbit user info to the session
    request.session['fitbit_profile'] = fb.user_profile_get()
    if utils.get_setting('FITAPP_SUBSCRIBE'):
        try:
            SUBSCRIBER_ID = utils.get_setting('FITAPP_SUBSCRIBER_ID')
        except ImproperlyConfigured:
            return redirect(reverse('fitbit-error'))
        subscribe.apply_async((fbuser.fitbit_user, SUBSCRIBER_ID), countdown=5)
        # Create tasks for all data in all data types
        for i, _type in enumerate(TimeSeriesDataType.objects.all()):
            # Delay execution for a few seconds to speed up response
            # Offset each call by 2 seconds so they don't bog down the server
            get_time_series_data.apply_async(
                (fbuser.fitbit_user, _type.category, _type.resource,),
                countdown=10 + (i * 5))

    next_url = request.session.pop('fitbit_next', None) or utils.get_setting(
        'FITAPP_LOGIN_REDIRECT')
    return redirect(next_url)


@receiver(user_logged_in)
def create_fitbit_session(sender, request, user, **kwargs):
    """ If the user is a fitbit user, update the profile in the session. """

    if user.is_authenticated() and utils.is_integrated(user) and \
            user.is_active:
        fbuser = UserFitbit.objects.filter(user=user)
        if fbuser.exists():
            fb = utils.create_fitbit(**fbuser[0].get_user_data())
            try:
                request.session['fitbit_profile'] = fb.user_profile_get()
            except:
                pass


@login_required
def error(request):
    """
    The user is redirected to this view if we encounter an error acquiring
    their Fitbit credentials. It renders the template defined in the setting
    :ref:`FITAPP_ERROR_TEMPLATE`. The default template, located at
    *fitapp/error.html*, simply informs the user of the error::

        <html>
            <head>
                <title>Fitbit Authentication Error</title>
            </head>
            <body>
                <h1>Fitbit Authentication Error</h1>

                <p>We encontered an error while attempting to authenticate you
                through Fitbit.</p>
            </body>
        </html>

    URL name:
        `fitbit-error`
    """
    return render(request, utils.get_setting('FITAPP_ERROR_TEMPLATE'), {})


@login_required
def logout(request):
    """Forget this user's Fitbit credentials.

    If the request has a `next` parameter, the user is redirected to that URL.
    Otherwise, they're redirected to the URL defined in the setting
    :ref:`FITAPP_LOGOUT_REDIRECT`.

    URL name:
        `fitbit-logout`
    """
    user = request.user
    try:
        fbuser = user.userfitbit
    except UserFitbit.DoesNotExist:
        pass
    else:
        if utils.get_setting('FITAPP_SUBSCRIBE'):
            try:
                SUBSCRIBER_ID = utils.get_setting('FITAPP_SUBSCRIBER_ID')
            except ImproperlyConfigured:
                return redirect(reverse('fitbit-error'))
            unsubscribe.apply_async(kwargs=fbuser.get_user_data(), countdown=5)
        fbuser.delete()
    next_url = request.GET.get('next', None) or utils.get_setting(
        'FITAPP_LOGOUT_REDIRECT')
    return redirect(next_url)


@csrf_exempt
def update(request):
    """Receive notification from Fitbit.

    Loop through the updates and create celery tasks to get the data.
    More information here:
    https://wiki.fitbit.com/display/API/Fitbit+Subscriptions+API

    URL name:
        `fitbit-update`
    """

    # The updates can come in two ways:
    # 1. A json body in a POST request
    # 2. A json file in a form POST
    if request.method == 'POST':
        try:
            body = request.body
            if request.FILES and 'updates' in request.FILES:
                body = request.FILES['updates'].read()
            updates = json.loads(body.decode('utf8'))
            # Create a celery task for each data type in the update
            for update in updates:
                cat = getattr(TimeSeriesDataType, update['collectionType'])
                resources = TimeSeriesDataType.objects.filter(category=cat)
                for i, _type in enumerate(resources):
                    # Offset each call by 2 seconds so they don't bog down the
                    # server
                    get_time_series_data.apply_async(
                        (update['ownerId'], _type.category, _type.resource,),
                        {'date': parser.parse(update['date'])},
                        countdown=(2 * i))
        except:
            return redirect(reverse('fitbit-error'))

        return HttpResponse(status=204)

    # if someone enters the url into the browser, raise a 404
    raise Http404


def make_response(code=None, objects=[]):
    """AJAX helper method to generate a response"""

    data = {
        'meta': {'total_count': len(objects), 'status_code': code},
        'objects': objects,
    }
    return HttpResponse(json.dumps(data))


def normalize_date_range(request, fitbit_data):
    """Prepare a fitbit date range for django database access. """

    result = {}
    base_date = fitbit_data['base_date']
    if base_date == 'today':
        now = timezone.now()
        if 'fitbit_profile' in request.session.keys():
            tz = request.session['fitbit_profile']['user']['timezone']
            now = timezone.pytz.timezone(tz).normalize(timezone.now())
        base_date = now.date().strftime('%Y-%m-%d')
    result['date__gte'] = base_date

    if 'end_date' in fitbit_data.keys():
        result['date__lte'] = fitbit_data['end_date']
    else:
        period = fitbit_data['period']
        if period != 'max':
            if type(base_date) in STRING_TYPES:
                start = parser.parse(base_date)
            else:
                start = base_date
            if 'y' in period:
                kwargs = {'years': int(period.replace('y', ''))}
            elif 'm' in period:
                kwargs = {'months': int(period.replace('m', ''))}
            elif 'w' in period:
                kwargs = {'weeks': int(period.replace('w', ''))}
            elif 'd' in period:
                kwargs = {'days': int(period.replace('d', ''))}
            end_date = start + relativedelta(**kwargs)
            result['date__lte'] = end_date.strftime('%Y-%m-%d')

    return result


@require_GET
def get_steps(request):
    """An AJAX view that retrieves this user's step data from Fitbit.

    This view has been deprecated. Use `get_data` instead.

    URL name:
        `fitbit-steps`
    """

    return get_data(request, 'activities', 'steps')


@require_GET
def get_data(request, category, resource):
    """An AJAX view that retrieves this user's data from Fitbit.

    This view may only be retrieved through a GET request. The view can
    retrieve data from either a range of dates, with specific start and end
    days, or from a time period ending on a specific date.

    The two parameters, category and resource, determine which type of data
    to retrieve. The category parameter can be one of: foods, activities,
    sleep, and body. It's the first part of the path in the items listed at
    https://wiki.fitbit.com/display/API/API-Get-Time-Series
    The resource parameter should be the rest of the path.

    To retrieve a specific time period, two GET parameters are used:

        :period: A string describing the time period, ending on *base_date*,
            for which to retrieve data - one of '1d', '7d', '30d', '1w', '1m',
            '3m', '6m', '1y', or 'max.
        :base_date: The last date (in the format 'yyyy-mm-dd') of the
            requested period. If not provided, then *base_date* is
            assumed to be today.

    To retrieve a range of dates, two GET parameters are used:

        :base_date: The first day of the range, in the format 'yyyy-mm-dd'.
        :end_date: The final day of the range, in the format 'yyyy-mm-dd'.

    The response body contains a JSON-encoded map with two items:

        :objects: an ordered list (from oldest to newest) of daily data
            for the requested period. Each day is of the format::

               {'dateTime': 'yyyy-mm-dd', 'value': 123}

           where the user has *value* on *dateTime*.
        :meta: a map containing two things: the *total_count* of objects, and
            the *status_code* of the response.

    When everything goes well, the *status_code* is 100 and the requested data
    is included. However, there are a number of things that can 'go wrong'
    with this call. For each type of error, we return an empty data list with
    a *status_code* to describe what went wrong on our end:

        :100: OK - Response contains JSON data.
        :101: User is not logged in.
        :102: User is not integrated with Fitbit.
        :103: Fitbit authentication credentials are invalid and have been
            removed.
        :104: Invalid input parameters. Either *period* or *end_date*, but not
            both, must be supplied. *period* should be one of [1d, 7d, 30d,
            1w, 1m, 3m, 6m, 1y, max], and dates should be of the format
            'yyyy-mm-dd'.
        :105: User exceeded the Fitbit limit of 150 calls/hour.
        :106: Fitbit error - please try again soon.

    See also the `Fitbit API doc for Get Time Series
    <https://wiki.fitbit.com/display/API/API-Get-Time-Series>`_.

    URL name:
        `fitbit-data`
    """

    # Manually check that user is logged in and integrated with Fitbit.
    user = request.user
    try:
        resource_type = TimeSeriesDataType.objects.get(
            category=getattr(TimeSeriesDataType, category), resource=resource)
    except:
        return make_response(104)

    fitapp_subscribe = utils.get_setting('FITAPP_SUBSCRIBE')
    if not user.is_authenticated() or not user.is_active:
        return make_response(101)
    if not fitapp_subscribe and not utils.is_integrated(user):
        return make_response(102)

    base_date = request.GET.get('base_date', None)
    period = request.GET.get('period', None)
    end_date = request.GET.get('end_date', None)
    if period and not end_date:
        form = forms.PeriodForm({'base_date': base_date, 'period': period})
    elif end_date and not period:
        form = forms.RangeForm({'base_date': base_date, 'end_date': end_date})
    else:
        # Either end_date or period, but not both, must be specified.
        return make_response(104)

    fitbit_data = form.get_fitbit_data()
    if not fitbit_data:
        return make_response(104)

    if fitapp_subscribe:
        # Get the data directly from the database.
        date_range = normalize_date_range(request, fitbit_data)
        existing_data = TimeSeriesData.objects.filter(
            user=user, resource_type=resource_type, **date_range)
        simplified_data = [{'value': d.value, 'dateTime': d.string_date()}
                           for d in existing_data]
        return make_response(100, simplified_data)

    # Request data through the API and handle related errors.
    fbuser = UserFitbit.objects.get(user=user)
    try:
        data = utils.get_fitbit_data(fbuser, resource_type, **fitbit_data)
    except (HTTPUnauthorized, HTTPForbidden):
        # Delete invalid credentials.
        fbuser.delete()
        return make_response(103)
    except HTTPConflict:
        return make_response(105)
    except HTTPServerError:
        return make_response(106)
    except:
        # Other documented exceptions include TypeError, ValueError,
        # HTTPNotFound, and HTTPBadRequest. But they shouldn't occur, so we'll
        # send a 500 and check it out.
        raise

    return make_response(100, data)



from setuptools import setup, find_packages

required = [line for line in open('requirements/base.txt').read().split("\n")]

setup(
    name="django-fitbit",
    version=__import__("fitapp").__version__,
    author="orcas",
    author_email="",
    packages=find_packages(),
    install_requires=["setuptools"] + required,
    include_package_data=True,
    url="https://github.com/orcasgit/django-fitbit/",
    license="",
    description="Django integration for python-fitbit",
    long_description=open("README.md").read(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        'License :: OSI Approved :: Apache Software License',
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 2 :: Only"
    ],
    test_suite="runtests.runtests"
)



# -*- coding: utf-8 -*-
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


# Safe User import for Django < 1.5
try:
    from django.contrib.auth import get_user_model
except ImportError:
    from django.contrib.auth.models import User
else:
    User = get_user_model()

# With the default User model these will be 'auth.User' and 'auth.user'
# so instead of using orm['auth.User'] we can use orm[user_orm_label]
user_orm_label = '%s.%s' % (User._meta.app_label, User._meta.object_name)
user_model_label = '%s.%s' % (User._meta.app_label, User._meta.module_name)


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding unique constraint on 'UserFitbit', fields ['fitbit_user']
        db.create_unique(u'fitapp_userfitbit', ['fitbit_user'])


    def backwards(self, orm):
        # Removing unique constraint on 'UserFitbit', fields ['fitbit_user']
        db.delete_unique(u'fitapp_userfitbit', ['fitbit_user'])


    models = {
        'auth.group': {
            'Meta': {'object_name': 'Group'},
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '80'}),
            'permissions': ('django.db.models.fields.related.ManyToManyField', [], {'to': "orm['auth.Permission']", 'symmetrical': 'False', 'blank': 'True'})
        },
        'auth.permission': {
            'Meta': {'ordering': "('content_type__app_label', 'content_type__model', 'codename')", 'unique_together': "(('content_type', 'codename'),)", 'object_name': 'Permission'},
            'codename': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'content_type': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['contenttypes.ContentType']"}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '50'})
        },
        user_model_label: {
            'Meta': {
                'object_name': User.__name__,
                'db_table': "'%s'" % User._meta.db_table
            },
            User._meta.pk.attname: (
                'django.db.models.fields.AutoField', [],
                {'primary_key': 'True',
                'db_column': "'%s'" % User._meta.pk.column}
            ),
        },
        'contenttypes.contenttype': {
            'Meta': {'ordering': "('name',)", 'unique_together': "(('app_label', 'model'),)", 'object_name': 'ContentType', 'db_table': "'django_content_type'"},
            'app_label': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'model': ('django.db.models.fields.CharField', [], {'max_length': '100'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '100'})
        },
        u'fitapp.timeseriesdata': {
            'Meta': {'unique_together': "(('user', 'resource_type', 'date'),)", 'object_name': 'TimeSeriesData'},
            'date': ('django.db.models.fields.DateField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'resource_type': ('django.db.models.fields.related.ForeignKey', [], {'to': u"orm['fitapp.TimeSeriesDataType']"}),
            'user': ('django.db.models.fields.related.ForeignKey', [], {'to': "orm['%s']" % user_orm_label}),
            'value': ('django.db.models.fields.CharField', [], {'default': 'None', 'max_length': '32', 'null': 'True'})
        },
        u'fitapp.timeseriesdatatype': {
            'Meta': {'ordering': "['category', 'resource']", 'unique_together': "(('category', 'resource'),)", 'object_name': 'TimeSeriesDataType'},
            'category': ('django.db.models.fields.IntegerField', [], {}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'resource': ('django.db.models.fields.CharField', [], {'max_length': '128'})
        },
        u'fitapp.userfitbit': {
            'Meta': {'object_name': 'UserFitbit'},
            'auth_secret': ('django.db.models.fields.TextField', [], {}),
            'auth_token': ('django.db.models.fields.TextField', [], {}),
            'fitbit_user': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '32'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'user': ('django.db.models.fields.related.OneToOneField', [], {'to': "orm['%s']" % user_orm_label, 'unique': 'True'})
        }
    }

    complete_apps = ['fitapp']



from fitapp.models import UserFitbit, TimeSeriesDataType
from django.db import IntegrityError

from .base import FitappTestBase


class TestFitappModels(FitappTestBase):
    def test_userfitbit(self):
        """ UserFitbit was already created in base, now test the properties """
        self.assertEqual(self.fbuser.user, self.user)
        self.assertEqual(self.fbuser.__str__(), self.username)
        self.assertEqual(self.fbuser.get_user_data(), {
            'resource_owner_key': self.fbuser.auth_token,
            'resource_owner_secret': self.fbuser.auth_secret,
            'user_id': self.fbuser.fitbit_user
        })

        # Trying to create another UserFitbit with the same fitbit_user should
        # result in an IntegrityError
        user2 = self.create_user(
            username='%s2' % self.username, password=self.password)
        self.assertRaises(IntegrityError, self.create_userfitbit,
                          user=user2, fitbit_user=self.fbuser.fitbit_user)

    def test_timeseriesdatatype(self):
        """ TimeSeriesDataTypes are created via fixtures. """
        self.assertEqual(TimeSeriesDataType.objects.count(), 36)
        assert hasattr(TimeSeriesDataType, 'activities')
        assert hasattr(TimeSeriesDataType, 'body')
        assert hasattr(TimeSeriesDataType, 'sleep')
        assert hasattr(TimeSeriesDataType, 'foods')
        self.assertEqual(str(TimeSeriesDataType.objects.get(resource='steps')),
                         'activities/steps')
